
/* BEGIN COPYRIGHT NOTICE

Copyright (c) 1992 by John Sarraille and Peter DiFalco
(john@ishi.csustan.edu)

Copyright (c) 1998 by John Sarraille
(john@ishi.csustan.edu)

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that copyright notice and this permission
notice appear in supporting documentation.

The algorithm used in this program was inspired by the paper
entitled "A Fast Algorithm To Determine Fractal Dimensions By
Box Counting", which was written by Liebovitch and Toth, and
which appeared in the journal "Physics Letters A", volume 141,
pp 386-390, (1989).

This program is not warranted: use at your own risk.

END COPYRIGHT NOTICE */


/*

   FD3 PROGRAM FILE

Filename:  file.c

OUTLINE OF CALLS:

Gfopen +
  fopen
  printf
  fflush
  exit
GfscanfDbl +
  fscanf
  printf
  fflush
  exit
GfscanfULI +
  fscanf
  printf
  fflush
  exit
PrintFileLines +
  CkBadParams +
    printf
    fflush
    exit
  Gfopen +
  ScanLines +
    ScanToEOLN +
    ScanOneEOLN +
  printf
  fflush
  fclose
SkipWhite +
  getc
  BackUpOneChar +
    fseek
    printf
    fflush
    exit
  TellStopChar +
    printf
    fflush
SkipSameLineWhite +
  getc
  BackUpOneChar +
  TellStopChar +
ScanToEOLN +
  getc
  printf
  BackUpOneChar +
  TellStopChar +
SkipEOLNchars +
  getc
  TestEOLNstop +
  BackUpOneChar +
  TellStopChar +
  MaxOfFour +
ScanOneEOLN +
  getc
  TestSOE +
  printf
  BackUpOneChar +
SkipBlack +
  getc
  IsWhite +
  BackUpOneChar +
  TellStopChar +
CountStringsOnLine +
  SkipSameLineWhite +
  SkipBlack +
GetNumFromDigits +
  InitGNFD +
  CkMULrep +
    printf
    fflush
  getc
  ProcessChar +
    GNFDreadREP +
      printf
      fflush
  BackUpOneChar +
  PrintGnfdRes +
     printf
     fflush
GetUlong +
  getc
  SynchOnFirstDig +
    BackUpOneChar +
  GetNumFromDigits +
*/

#include "fd.h"

/*
     THE FUNCTIONS IN THIS FILE:
*/

void  Gfopen              () ;
void  GfscanfDbl          () ;
void  GfscanfULI          () ;
void  PrintFileLines      () ;
void  CkBadParams         () ;
int   ScanLines           () ;
int   SkipWhite           () ;
void  BackUpOneChar       () ;
void  TellStopChar        () ;
int   SkipSameLineWhite   () ;
int   ScanToEOLN          () ;
int   SkipEOLNchars       () ;
int   TestEOLNstop        () ;
int   MaxOfFour           () ;
int   ScanOneEOLN         () ;
int   TestSOE             () ; 
int   SkipBlack           () ;
int   IsWhite             () ;
int   CountStringsOnLine  () ;
int   GetNumFromDigits    () ;
void  InitGNFD            () ;
void  CkMULrep            () ;
int   ProcessChar         () ;
void  GNFDreadREP         () ;
void  PrintGnfdRes        () ;
int   GetUlong            () ;
int   SynchOnFirstDig     () ;

/* ######################################################### */
/*                        Gfopen                             */
/* ######################################################### */
/*
       FUNCTION TO GRACEFULLY OPEN A FILE.
*/      
void Gfopen (caller, fileVarPtrPtr, fileName, modeString)
     char     caller     [] ;
     FILE **  fileVarPtrPtr    ;
     char     fileName   [] ;
     char *   modeString ;
{
   if( ( *fileVarPtrPtr = fopen(fileName, modeString)) == NULL  )
   {
     printf("function %s was not able to ", caller ) ;
     printf("open the file named: %s. \n\n", fileName);
     fflush(stdout); exit(EXIT_FAILURE);
   }
   return ;
}
/* ######################################################### */
/*                      GfscanfDbl                           */
/* ######################################################### */
/*
         FUNCTION TO GRACEFULLY fscanf A double FROM A FILE.

         NOTE: TESTS ON MY SYSTEM INDICATE THAT fscanf DOES
         *NOT* RETURN ZERO OR EOF WHEN THE INPUT IS LARGER THAN
         DBL_MAX.  THE VALUE OF ERRNO DOES NOT HELP EITHER.

         FOR NOW, I AM JUST CHECKING TO SEE IF THE VALUE READ
         IS GREATER THAN THE MAXIMUM VALUE OF A DOUBLE, AND IF
         IT IS LESS THAN THE NEGATIVE OF THAT.  IF THE COMPILER
         USED SUPPORTS THE VALUES +/- Infinity, THEN THIS TEST
         MAY BE SUFFICIENT FOR CHECKING FOR OVERFLOW.  (IT
         SEEMS TO BE WORKING PROPERLY ON MY SYSTEM.) IN ANY
         CASE, I DON'T SEE ANY HARM IN DOING THE TEST.  PEOPLE
         CAN TAKE IT OUT IF IT CAUSES PROBLEMS.  

         THIS PROCEDURE SHOULD BE RE-WRITTEN SOME TIME.  WE
         SHOULD HAVE A PROCEDURE THAT INTERPRETS THE DATA A
         CHARACTER AT A TIME.
         
*/
void GfscanfDbl (caller, fileName, lineNum, fileVarPtr,
                   dblVarPtr, maxDouble )
     char     caller [], fileName [] ;
     uLong    lineNum  ;
     FILE   * fileVarPtr    ;
     double * dblVarPtr     ;
     double   maxDouble     ;
{
  int scanRslt ;

  scanRslt = fscanf(fileVarPtr, "%lf" , dblVarPtr);
/* ADD THIS CODE IF YOU WANT TO CHECK SCAN RESULTS CLOSELY.
  printf("\n") ;
  printf("scanRslt is: %d\n",scanRslt) ;
  printf("errno is: %d\n", errno) ;
  printf("(* dblVarPtr) is: %f\n",(* dblVarPtr)) ;
  printf("\n") ;
*/
  if ( (scanRslt == 0) || (scanRslt == EOF)
        || ((* dblVarPtr) > maxDouble)
        || ((* dblVarPtr) < -maxDouble)
     )
     { printf ("\n%s: Bad format, bad value,", caller) ;
       printf ("or premature end of file \n") ;     
       printf ("on line %lu of file %s.\n\n" , lineNum, fileName) ;
       fflush(stdout); exit (EXIT_FAILURE) ; }
  return ;
}
/* ######################################################### */
/*                      GfscanfULI                           */
/* ######################################################### */
/*
        FUNCTION TO GRACEFULLY fscanf A uLong FROM A FILE.

        ALTHOUGH I ATTEMPTED TO PROGRAM IN A CHECK, ON MY
        SYSTEM, THIS FUNCTION *FAILS* TO NOTICE WHEN IT READS
        AN INTEGER BIGGER THAN THE MAX VALUE FOR AN UNSIGNED
        LONG.

        SEE THE COMMENTS ACCOMPANYING GfscanDbl.  THERE *IS* A
        CHARACTER-AT-A-TIME VERSION OF THIS FUNCTION, GetUlong,
        IN THIS FILE.
*/
void GfscanfULI (caller, fileName, lineNum, fileVarPtr,
                   uliVarPtr, maxUlong )
     char     caller   [] ;     char     fileName [] ;
     uLong    lineNum     ;     uLong  * uliVarPtr   ;
     FILE   * fileVarPtr  ;     uLong    maxUlong    ;
{
  int scanRslt ;

  scanRslt = fscanf(fileVarPtr, "%lu" , uliVarPtr);

  if ( (scanRslt == 0)
         || (scanRslt == EOF)
         || (* uliVarPtr) > maxUlong)
     { printf ("\n%s: Bad format, bad value,", caller) ;
       printf ("or premature end of file \n") ;     
       printf ("on line %lu of file %s.\n\n" , lineNum, fileName) ;
       fflush(stdout); exit (EXIT_FAILURE) ; }
  return ;
}
/* ######################################################### */
/*                     PrintFileLines                        */
/* ######################################################### */
/*
    THIS FUNCTION OPENS THE INDICATED TEXT FILE AND PRINTS THE
    INDICATED LINES, OR AS MANY OF THEM AS EXIST.

     DEVELOPER NOTES:

     DON'T CHANGE THE "rb" FILE MODE PARAMETER IN THE Gfopen
     CALL, OR ELSE YOU RISK SEEKS FAILING ON DOS SYSTEMS.
*/
void PrintFileLines (commandPtr, fileName, startLineNum, stopLineNum)
       commandType * commandPtr ;
       char        * fileName ;
       uLong startLineNum, stopLineNum ;
{ 
  FILE    * filePtr ;
  int     eofStop ;
  uLong   howMany ;

  CkBadParams (startLineNum, stopLineNum) ;
  Gfopen("PrintFileLines", &filePtr, fileName, "rb");
  howMany = startLineNum - 1 ;
  eofStop = ScanLines (commandPtr, filePtr, howMany, FALSE) ;
  howMany = stopLineNum - startLineNum + 1 ;
  if ( !eofStop)
     eofStop = ScanLines (commandPtr, filePtr, howMany, TRUE) ;  
  printf("\n") ; 
  fflush(stdout) ;
  fclose (filePtr) ;
  return ;
}
/* ######################################################### */
/*                       CkBadParams                         */
/* ######################################################### */
void   CkBadParams (startLineNum, stopLineNum)
         uLong startLineNum, stopLineNum ;
{
   if ( (startLineNum < 1) || (stopLineNum < startLineNum) )
      {
         printf("Bad startLineNum or stopLineNum parameter to ");
         printf("function PrintFileLines.\n\n");
         fflush(stdout) ; exit(EXIT_FAILURE);
      }

   return ;
}
/* ######################################################### */
/*                        ScanLines                          */
/* ######################################################### */
/*
      SCAN howMany LINES.  ECHO THEM IF willEcho IS TRUE.
*/
int  ScanLines (commandPtr, filePtr, howMany, willEcho ) 
       commandType * commandPtr ;
       FILE        * filePtr    ;
       uLong         howMany    ;
       int           willEcho   ;
{
   int    eofStop       = FALSE ;
   uLong  linesScanned  =     0 ;

   while (!eofStop && (linesScanned < howMany))
   { 
      eofStop=(ScanToEOLN(commandPtr,filePtr,willEcho) == EOF) ;
      if (!eofStop)
      { 
        eofStop=(ScanOneEOLN(commandPtr,filePtr,willEcho) == EOF) ;
        linesScanned++ ;
      }
   }
   return (eofStop) ;
}
/* ######################################################### */
/*                        SkipWhite                          */
/* ######################################################### */
/*
       FROM THE CURRENT FILE POSITION, THIS FUNCTIOIN READS ONE
       CHARACTER AT A TIME UNTIL A "BLACK SPACE" CHARACTER (A
       CHARACTER THAT IS NOT WHITE SPACE) IS READ, OR UNTIL END
       OF FILE IS REACHED.  WE AVOID INTERPRETING nextCh AS A
       CHARACTER, EXCEPT IN CASES WHERE WE KNOW nextCh != EOF.
       
       IF A "BLACK SPACE" CHARACTER IS ENCOUNTERED, THEN WE
       "BACK UP" SO IT WILL BE THE NEXT CHARACTER READ.  WE
       ARRANGE TO TRAP A FAILURE IN THE SEEK COMMAND AND ABORT
       THE PROGRAM IF DEBUGGING IS TURNED ON.

       THIS FUNCTION RETURNS THE VALUE RETURNED BY THE LAST CALL
       TO getc.  IF THE INTEGER VALUE RETURNED IS NOT "EOF",
       THEN THE (char) CONVERSION OF THIS INTEGER WILL BE THE
       LAST CHARACTER READ BY SkipWhite.  AS LONG AS THE SEEK
       IS SUCCESSFUL, THE CHARACTER IS "PUT BACK" SO IT WILL BE
       THE NEXT CHARACTER READ.
*/
int SkipWhite (commandPtr, filePtr)
       commandType * commandPtr ;
       FILE * filePtr ;
{
  int nextCh, stop = FALSE ;

  do
  { nextCh = getc (filePtr) ;
    if ( nextCh == EOF )   stop = TRUE ;
    else 
       stop = (     (((char) nextCh) != '\r') && (((char) nextCh) != '\f')
                 && (((char) nextCh) != '\t') && (((char) nextCh) != '\n')
                 && (((char) nextCh) != '\v') && (((char) nextCh) != ' ' ));
  }
  while (!stop) ;

  if (nextCh != EOF) BackUpOneChar (commandPtr, filePtr, "SkipWhite") ;
  if ( (commandPtr->debugging) && (commandPtr->verbose) )
     TellStopChar (nextCh, "SkipWhite") ;

  return (nextCh) ;
}
/* ######################################################### */
/*                     BackUpOneChar                         */
/* ######################################################### */
void BackUpOneChar (commandPtr, filePtr, caller)
       commandType * commandPtr ;
       FILE        * filePtr ;
       char        * caller ;
{
   int seekFailed ;
         
   seekFailed = fseek(filePtr, -1L, SEEK_CUR) ;
   if (  seekFailed && (commandPtr->debugging)  )
   {
      printf("fseek failure in function \"%s\". \n\n", caller) ;
      fflush(stdout); exit(EXIT_FAILURE) ;
   }

   return ;
}
/* ######################################################### */
/*                     TellStopChar                          */
/* ######################################################### */
void  TellStopChar (stopCh, caller)
          int    stopCh ;
          char   * caller ;
{
  printf ("\"%s\" ran and stopped on the character", caller ) ;
  if ((stopCh >= 32) && (stopCh <= 126) )
        printf (": \"%c\"\n\n" , ((char) stopCh)) ;
  else  printf (" with decimal number: \"%d\"\n\n" , ((char) stopCh)) ;
  fflush(stdout);  

  return ;
}
/* ######################################################### */
/*                   SkipSameLineWhite                       */
/* ######################################################### */
/*
       FROM THE CURRENT FILE POSITION, THIS FUNCTION READS ONE
       CHARACTER AT A TIME UNTIL A "BLACK SPACE" CHARACTER (A
       CHARACTER THAT IS NOT WHITE SPACE) IS READ, OR UNTIL END
       OF FILE IS REACHED, OR UNTIL END OF LINE IS REACHED.  WE
       AVOID INTERPRETING nextCh AS A CHARACTER, EXCEPT IN CASES
       WHERE WE KNOW nextCh != EOF.
       
       IF A BLACK SPACE CHARACTER OR END-OF-LINE IS
       ENCOUNTERED, THE FUNCTION WILL "BACK UP" SO THAT
       CHARACTER WILL BE THE NEXT ONE READ.  WE ARRANGE TO TRAP
       A FAILURE IN THE SEEK COMMAND AND ABORT THE PROGRAM IF
       DEBUGGING IS TURNED ON.

       THE FUNCTION RETURNS THE VALUE RETURNED BY THE LAST CALL
       TO getc.  IF THE INTEGER VALUE RETURNED IS NOT "EOF",
       THEN THE (char) CONVERSION OF THIS INTEGER WILL BE THE
       LAST CHARACTER READ BY SkipSameLineWhite.  AS LONG AS
       THE SEEK IS SUCCESSFUL, THE CHARACTER IS "PUT BACK" SO
       IT WILL ALSO BE THE NEXT CHARACTER READ.
*/
int SkipSameLineWhite (commandPtr, filePtr)
       commandType * commandPtr ;
       FILE * filePtr ;
{
  int nextCh, stop = FALSE ;
  do
  { nextCh = getc (filePtr) ;
    if (nextCh == EOF )
       stop = TRUE ;
    else 
       stop = (   (((char) nextCh) != '\t')
               && (((char) nextCh) != ' ' ) ) ;
  }
  while (!stop) ;
  if (nextCh != EOF)
     BackUpOneChar (commandPtr, filePtr, "SkipSameLineWhite") ;

  if ( (commandPtr->debugging) && (commandPtr->verbose) )
     TellStopChar (nextCh, "SkipSameLineWhite") ;

  return (nextCh) ;
}
/* ######################################################### */
/*                       ScanToEOLN                          */
/* ######################################################### */
/*
       FROM THE CURRENT FILE POSITION, READ ONE CHARACTER AT A
       TIME UNTIL END OF LINE IS REACHED, OR UNTIL END OF FILE
       IS REACHED.  IF THE PARAMETER willEcho IS "TRUE" THIS
       FUNCTION WRITES TO stdout EACH CHARACTER THAT IT READS,
       EXCEPT THAT USUALLY IT DOES NOT WRITE THE LAST CHARACTER
       READ -- THE "STOP CHARACTER".  IN THE SPECIAL CASE WHERE
       (PREMATURE) EOF IS ENCOUNTERED AND willEcho IS TRUE, ALL
       CHARACTERS READ ARE WRITTEN.  WE AVOID INTERPRETING
       nextCh AS A CHARACTER, EXCEPT IN CASES WHERE WE KNOW
       nextCh != EOF.

       IF AN END-OF-LINE CHARACTER IS ENCOUNTERED, THEN WE BACK
       UP SO IT WILL BE THE NEXT CHARACTER READ.  WE ARRANGE TO
       TRAP A FAILURE IN THE SEEK COMMAND AND ABORT THE PROGRAM
       IF DEBUGGING IS TURNED ON.

       THE FUNCTION RETURNS THE VALUE RETURNED BY THE LAST CALL
       TO getc.  IF THE INTEGER VALUE RETURNED IS NOT "EOF",
       THEN THE (char) CONVERSION OF THIS INTEGER WILL BE THE
       LAST CHARACTER READ BY ScanToEOLN.  AS LONG AS THE SEEK
       IS SUCCESSFUL, THE CHARACTER IS "PUT BACK" SO IT WILL
       ALSO BE THE NEXT CHARACTER READ.
*/
int ScanToEOLN (commandPtr, filePtr, willEcho)
       commandType * commandPtr ;
       FILE        * filePtr ;
       int           willEcho ;
{
  int nextCh, stop = FALSE ;
  do
  {
    nextCh = getc (filePtr) ;
    if (nextCh == EOF ) stop = TRUE ;
    else 
       stop = (    (((char) nextCh) == '\r') || (((char) nextCh) == '\f')
                || (((char) nextCh) == '\n') || (((char) nextCh) == '\v')) ;
    if ( (!stop) && willEcho ) printf("%c", ( (char) nextCh )) ;
  }
  while (!stop) ;

  if (nextCh != EOF)  BackUpOneChar (commandPtr, filePtr, "ScanToEOLN") ;
  if ( (commandPtr->debugging) && (commandPtr->verbose) )
     TellStopChar (nextCh, "ScanToEOLN") ;

  return (nextCh) ;
}
/* ######################################################### */
/*                     SkipEOLNchars                         */
/* ######################################################### */
/*
     FROM THE CURRENT FILE POSITION, THIS FUNCTION READS A
     CHARACTER AT A TIME UNTIL HITTING EOF OR ANY CHARACTER
     THAT IS NOT A return, formfeed, newline OR vertical tab.
     IT RETURNS AN ESTIMATE OF THE NUMBER OF LINES IT HAS GONE
     PAST.  THIS ESTIMATE IS THE MAXIMUM OF THE NUMBERS OF
     return's, form feed's, newline's, AND vertical tab's THAT
     HAVE BEEN READ.

     IF A BLACK SPACE CHARACTER, space or, horizontal tab IS
     ENCOUNTERED, THE FUNCTION "BACKS UP" SO THIS CHARACTER
     WILL BE THE NEXT ONE READ.  WE ARRANGE TO TRAP A FAILURE
     IN THE SEEK COMMAND AND ABORT THE PROGRAM IF DEBUGGING IS
     TURNED ON.
*/
int SkipEOLNchars (commandPtr, filePtr)
       commandType * commandPtr ;
       FILE * filePtr ;
{
  int nextCh, stop = FALSE, returns = 0, feeds = 0,
      newlines = 0, vTabs = 0 ;
  do
  {
    nextCh = getc (filePtr) ;
    if    (nextCh == EOF)
          stop = TRUE ;
    else  stop = TestEOLNstop (nextCh, &returns, &feeds, &newlines, &vTabs) ;
  }
  while (!stop) ;
  if (nextCh != EOF) BackUpOneChar (commandPtr,filePtr,"SkipEOLNchars") ;

  if ( (commandPtr->debugging) && (commandPtr->verbose) )
     TellStopChar (nextCh, "SkipEOLNchars") ;

  return( MaxOfFour (returns, feeds, newlines, vTabs) ) ;    
}
/* ######################################################### */
/*                       TestEOLNstop                        */
/* ######################################################### */
/*
     INCREMENT THE APPROPRIATE COUNTER IF THE INPUT CHARACTER
     IS AN END-OF-LINE CHARACTER, ELSE SET A FLAG TO STOP
     READING.
*/
int TestEOLNstop (nextCh, retPtr, fdPtr, nlPtr, vTptr) 
      int nextCh, * retPtr, * fdPtr, * nlPtr, * vTptr  ;
{
   int stop ;
   
   if (  ((char) nextCh) == '\r' ) (*retPtr)++  ;
   if (  ((char) nextCh) == '\f' ) (*fdPtr)++   ;
   if (  ((char) nextCh) == '\n' ) (*nlPtr)++   ;
   if (  ((char) nextCh) == '\v' ) (*vTptr)++   ;
   stop =      (((char) nextCh) != '\r')
            && (((char) nextCh) != '\f') 
            && (((char) nextCh) != '\n')
            && (((char) nextCh) != '\v') ;
    
    return (stop) ;
}
/* ######################################################### */
/*                       MaxOfFour                           */
/* ######################################################### */
int MaxOfFour (a, b, c, d) 
     int a, b, c, d ;
{
   int max = a ;
   if (b > max) max = b ;
   if (c > max) max = c ;
   if (d > max) max = d ;
   
   return (max) ;
}
/* ######################################################### */
/*                        ScanOneEOLN                        */
/* ######################################################### */
/*
     THIS FUNCTION ATTEMPTS TO GO PAST ONE AND ONLY ONE LINE
     TERMINATOR.  IT IS SUPPOSED TO BE USED WHEN THE NEXT
     CHARACTER TO BE READ IS A newline, carriage return, OR line
     feed.  IT ATTEMPTS TO SIMPLY MOVE TO THE BEGINNING OF THE
     NEXT LINE.  IT IS DOUBTLESS AN EXAMPLE OF
     OVER-ENGINEERING, BUT IT WILL WORK O.K. IN THE STANDARD
     ENVIRONMENTS, AND MAYBE IN A FEW THAT ARE UNUSUAL.  SINCE
     THIS FUNCTION IS NOT CALLED UPON TO PERFORM ANY
     "MISSION CRITICAL" TASKS, THERE'S NO RISK TO THE
     CORRECTNESS OF FD3 PROGRAM RESULTS IF THIS FUNCTION
     DOESN'T WORK PERFECTLY.  WHEN THE PARAMETER willEcho IS
     "TRUE", THE FUNCTION WRITES THE CHARACTERS COMPOSING THE
     LINE TERMINATOR.  THE FUNCTION ATTEMPTS *NOT* TO WRITE
     ANYTHING FOLLOWING THE FIRST LINE TERMINATOR.

     WE STOP READING IF WE HAVE COME TO A CHARACTER THAT IS NOT
     USED AS A LINE TERMINATOR, OR IF WE HAVE ALREADY GONE PAST
     A LINE TERMINATOR AND WE SEEM TO BE ENCOUNTERING ANOTHER
     ONE.  LINE TERMINATORS VARY FROM SYSTEM TO SYSTEM.
     PROBABLY MOST COMPILERS TAKE CARE OF TRANSLATING THEM SO
     THE C LANGUAGE "SEES" THEM AS newlines.  IN ANY CASE, WE
     ARRANGE THINGS SO THAT A return, line feed, newline,
     vertical tab OR COMBINATION OF return AND line feed WILL
     ALL BE INTERPRETED AS ONE SINGLE LINE TERMINATOR.
     
     WHEN THERE APPEARS TO BE SOMETHING TO READ ON THE NEXT
     LINE, THIS FUNCTION "BACKS UP" SO THE NEXT READ WILL GET
     THE FIRST CHARACTER ON THE NEXT LINE.  WE ARRANGE TO TRAP A
     FAILURE IN THE SEEK COMMAND AND ABORT THE PROGRAM IF
     DEBUGGING IS TURNED ON.
*/
int ScanOneEOLN (commandPtr, filePtr, willEcho)
       commandType * commandPtr ;
       FILE        * filePtr ;
       int           willEcho ;
{
  int nextCh, stop = FALSE, returns = 0, feeds = 0,
      newlines = 0, vTabs = 0 ;
  do
  {
    nextCh = getc (filePtr) ;
    if (nextCh == EOF)  stop = TRUE ;
    else stop = TestSOE (nextCh, &returns, &feeds, &newlines, &vTabs) ;
    if ( (!stop) && willEcho ) printf("%c", ( (char) nextCh )) ;
  }
  while (!stop) ;
  if (nextCh != EOF) BackUpOneChar (commandPtr, filePtr, "ScanOneEOLN") ;
  return(nextCh) ;  
}
/* ######################################################### */
/*                          TestSOE                          */
/* ######################################################### */
/*
     TEST TO SEE IF WE HAVE ALREADY GONE PAST A LINE TERMINATOR
     AND WE SEEM TO BE ENCOUNTERING ANOTHER ONE.
*/
int TestSOE (nextCh, retPtr, fdPtr, nlPtr, vTptr) 
      int nextCh, * retPtr, * fdPtr, * nlPtr, * vTptr  ;
{
   int stop ;
   if (((char)nextCh)=='\r') (* retPtr)++;
   if (((char)nextCh)=='\f') (* fdPtr)++;
   if (((char)nextCh)=='\n') (* nlPtr)++;
   if (((char)nextCh)=='\v') (* vTptr)++;
   stop =  (     (((char) nextCh) != '\r')
              && (((char) nextCh) != '\f')
              && (((char) nextCh) != '\n')
              && (((char) nextCh) != '\v')
           )
           || ((* retPtr) > 1)
           || ((* fdPtr) > 1)
           || ((* nlPtr) > 1)
           || ((* vTptr)   > 1)
           || ( ((* retPtr) + (* fdPtr) + (* nlPtr) + (* vTptr)) > 2 );
   
   return (stop) ;
}
/* ######################################################### */
/*                       SkipBlack                           */
/* ######################################################### */
/*
   FROM THE CURRENT FILE POSITION THIS FUNCTION READS ONE
   CHARACTER AT A TIME UNTIL A WHITE SPACE CHARACTER IS READ,
   OR UNTIL END OF FILE IS REACHED.  WE AVOID INTERPRETING
   nextCh AS A CHARACTER, EXCEPT IN CASES WHERE WE KNOW nextCh
   != EOF.

   IF A WHITE SPACE CHARACTER IS ENCOUNTERED, THE FUNCTION
   "BACKS UP" SO THIS CHARACTER WILL BE THE NEXT ONE READ.  WE
   ARRANGE TO TRAP A FAILURE IN THE SEEK COMMAND AND ABORT THE
   PROGRAM IF DEBUGGING IS TURNED ON.

   THE FUNCTION RETURNS THE VALUE RETURNED BY THE LAST CALL TO
   getc.  IF THE INTEGER VALUE RETURNED IS NOT "EOF", THEN THE
   (char) CONVERSION OF THIS INTEGER WILL BE THE LAST CHARACTER
   READ BY SkipBlack.  AS LONG AS THE SEEK IS SUCCESSFUL, THE
   CHARACTER IS "PUT BACK" SO IT WILL ALSO BE THE NEXT
   CHARACTER READ.
*/
int SkipBlack (commandPtr, filePtr)
       commandType * commandPtr ;
       FILE * filePtr ;
{
  int nextCh, stop = FALSE ;
  
  do
  {
    nextCh = getc (filePtr) ;
    if   ( nextCh == EOF )
         stop = TRUE ;
    else stop = IsWhite (nextCh) ;
  }
  while (!stop) ;

  if (nextCh != EOF) BackUpOneChar (commandPtr, filePtr, "SkipBlack") ;
  if ( (commandPtr->debugging) && (commandPtr->verbose) )
     TellStopChar (nextCh, "SkipBlack") ;

  return (nextCh) ;
}
/* ######################################################### */
/*                         IsWhite                           */
/* ######################################################### */
int IsWhite (theChar)
      int theChar ;
{
   return (       (((char) theChar) == '\r')
               || (((char) theChar) == '\f')
               || (((char) theChar) == '\t')
               || (((char) theChar) == '\n')
               || (((char) theChar) == '\v')
               || (((char) theChar) ==  ' ')
           )  ;
}
/* ######################################################### */
/*                    CountStringsOnLine                     */
/* ######################################################### */
/*
     COUNT THE NUMBER OF STRINGS ON THE CURRENT LINE, FROM THE
     CURRENT POSITION UP TO THE END OF THE LINE OR END-OF-FILE,
     IF IT COMES FIRST.
*/
int CountStringsOnLine (commandPtr, filePtr)
       commandType * commandPtr ;
       FILE * filePtr ;
{ int nextCh, stop = FALSE, count = 0 ;
  do
  {
    nextCh = SkipSameLineWhite (commandPtr, filePtr) ;
    if (nextCh == EOF)
       stop = TRUE ;
    else
       stop =     (((char) nextCh) == '\r')
               || (((char) nextCh) == '\f')
               || (((char) nextCh) == '\n')
               || (((char) nextCh) == '\v') ;
     if (!stop)
        { count++ ;
          SkipBlack (commandPtr, filePtr); }
  }
  while (!stop) ;
  return (count) ;
}
/* ######################################################### */
/*                      numPropsType                         */
/* ######################################################### */
typedef struct
{
   int digChr, digNum, numDigits, overflow ;
}  numPropsType ;
/* ######################################################### */
/*                     GetNumFromDigits                      */
/* ######################################################### */
/*
        THIS FUNCTION WILL READ A SEQUENCE OF DIGITS AND
        INTERPRET IT AS AN UNSIGNED LONG INTEGER.  IT READS ONE
        CHARACTER AT A TIME UNTIL ENCOUNTERING A NON-DIGIT, OR
        END-OF-FILE.  THE FOLLOWING ERRORS ARE TRAPPED: 1) NO
        DIGITS ARE FOUND BEFORE THE FIRST NON-DIGIT, AND 2) THE
        NUMBER REPRESENTED BY THE SEQUENCE OF DIGITS READ
        OVERFLOWS THE DATA TYPE UNSIGNED LONG INT.

        WE DON'T STOP READING DIGITS IF OVERFLOW OCCURS.  WE
        WANT THE SCAN TO GET PAST THE STRING OF DIGITS.

        IF A WHITE SPACE CHARACTER IS ENCOUNTERED, WE "BACK UP"
        SO IT WILL BE THE NEXT CHARACTER READ.  WE ARRANGE TO
        TRAP A FAILURE IN THE SEEK COMMAND AND ABORT THE
        PROGRAM IF DEBUGGING IS TURNED ON.

        THE FUNCTION RETURNS "TRUE" OR "FALSE" DEPENDING ON
        WHETHER THE CONVERSION IS SUCCESSFUL OR UNSUCESSFUL.
*/
int GetNumFromDigits(cmdPtr, fPtr, maxUlong, valuePtr)
       commandType        *  cmdPtr ;
       FILE               *  fPtr ;
       uLong                 maxUlong, *  valuePtr ;
{
  numPropsType  nP ;
  int   stop = FALSE ;

  InitGNFD (valuePtr, &nP) ;
  CkMULrep (cmdPtr, maxUlong) ;
  do
  {
    nP.digChr = getc(fPtr) ;
    if    (nP.digChr == EOF)   stop = TRUE ;
    else  stop = ProcessChar (cmdPtr, &nP, valuePtr, maxUlong) ;
  }
  while (!stop) ;
  if ((nP.digChr) != EOF) BackUpOneChar (cmdPtr,fPtr,"GetNumFromDigits") ;
  if (cmdPtr->debugging) PrintGnfdRes (* valuePtr, &nP) ;

  if  ( (nP.numDigits == 0) || (nP.overflow) )  return (FALSE) ;
  else return (TRUE) ;
}
/* ######################################################### */
/*                        InitGNFD                           */
/* ######################################################### */
void  InitGNFD (valuePtr, nPt) 
       uLong              *  valuePtr ;
       numPropsType       *  nPt ;
{
  (* valuePtr) = 0 ;    
  (nPt->numDigits) = 0 ;
  (nPt->overflow) = FALSE ;

   return ;
}

/* ######################################################### */
/*                         CkMULrep                          */
/* ######################################################### */
void  CkMULrep (commandPtr, maxUlong) 
       commandType        *  commandPtr ;
       uLong                 maxUlong ;
{
  if (commandPtr->debugging)
     {
       printf("maxUlong is: %lu\n\n",maxUlong);
       fflush(stdout);
     }  

   return ;
}
/* ######################################################### */
/*                       ProcessChar                         */
/* ######################################################### */
int ProcessChar (commandPtr, npP, valuePtr, maxUlong)
       commandType        *  commandPtr ;
       numPropsType       *  npP ;
       uLong              *  valuePtr, maxUlong ;
{
   int stop  =  FALSE ;

   if ((commandPtr->debugging) && (commandPtr->verbose)) GNFDreadREP (npP) ;

   if    ((((char) (npP->digChr) ) < '0') || (((char) (npP->digChr)) > '9'))
         stop = TRUE ;
   else
   {
     (npP->numDigits)++ ;
     (npP->digNum) = (int) (((char) (npP->digChr)) - '0') ;
     if (!(npP->overflow))
     { if   ((* valuePtr) > ((maxUlong - (npP->digNum))/10.0 ))
            (npP->overflow) = TRUE ;
       else (* valuePtr) = ( (* valuePtr) * 10 ) + (npP->digNum) ; }
   }
   return (stop) ;
}
/* ######################################################### */
/*                       GNFDreadREP                         */
/* ######################################################### */
void GNFDreadREP (npP)
       numPropsType       *  npP ;
{
   printf ("Loop in GetNumFromDigits ") ;
   printf ("just read: %c. \n", (char) (npP->digChr) ) ;
   fflush(stdout);

   return ;
} 
/* ######################################################### */
/*                       PrintGnfdRes                        */
/* ######################################################### */
void  PrintGnfdRes ( value, npP)
         uLong             value ;
         numPropsType   *  npP ;
{
   printf ("function \"GetNumFromDigits\" got the following\n") ;
   printf ("value for variable dataLines: %lu \n", value) ;
   printf ("Variable numDigits is: %d\n", npP->numDigits);
   printf ("Variable overflow is: ");
   if   (npP->overflow)
        printf ("TRUE.\n\n");
   else printf("FALSE.\n\n");
   fflush(stdout);

   return ;
}
/* ######################################################### */
/*                          GetUlong                         */
/* ######################################################### */
/*
   THIS FUNCTION READS A STRING.  IF THE STRING REPRESENTS A
   NON-NEGATIVE INTEGER THAT WILL FIT IN THE SIZE OF AN
   UNSIGNED LONG INT, THEN THE FUNCTION "RETURNS" THAT VALUE BY
   PUTTING IT IN THE LOCATION POINTED TO BY valuePtr .  IN THIS
   CASE THE ACTUAL RETURN VALUE OF THE FUNCTION IS "TRUE".
   OTHERWISE, THE FUNCTION RETURNS "FALSE".  IT IS AN ERROR IF
   THIS FUNCTION IS CALLED WHEN THERE IS WHITE SPACE STILL TO
   BE READ BEFORE THE START OF THE STRING REPRESENTING THE
   INTEGER. AN OPTIONAL LEADING PLUS SIGN IS ALLOWED, BUT
   OTHERWISE THE STRING MUST BE A SIMPLE SEQUENCE OF DIGITS, OR
   IT IS CONSIDERED AN ERROR.

   IF THE FIRST CHARACTER IN THE STRING IS AN INTEGER THE
   FUNCTION "BACKS UP" SO THE NEXT FUNCTION CALL CAN START TO
   READ THE DIGITS FROM THE BEGINNING.  WE ARRANGE TO TRAP A
   FAILURE IN THE SEEK COMMAND AND ABORT THE PROGRAM IF
   DEBUGGING IS TURNED ON.

   THERE IS A FUNCTION, GfscanfULI, IN THIS FILE.  IT ALSO
   READS AN UNSIGNED LONG INT.  GfscanfULI USES THE fscanf
   FUNCTION, SO IT IS PROBABLY NOT AS ROBUST AS THIS ONE.

*/
int GetUlong(commandPtr, filePtr, maxUlong, valuePtr)
       commandType  *  commandPtr ;
       FILE         *  filePtr    ;
       uLong           maxUlong   ;
       uLong        *  valuePtr   ;
{ 
  int curCh, chIsGood ;

  (* valuePtr) = 0 ;
  if (  (curCh = getc (filePtr)) == EOF ) return (FALSE) ;
  chIsGood = SynchOnFirstDig (commandPtr, filePtr, curCh ) ;
  return chIsGood &&
          (GetNumFromDigits (commandPtr, filePtr, maxUlong, valuePtr)) ;
}
/* ######################################################### */
/*                     SynchOnFirstDig                       */
/* ######################################################### */
int SynchOnFirstDig (cPtr, fPtr, curCh) 
       commandType        *  cPtr ;
       FILE               *  fPtr ;
       int                   curCh ;
{
   int chIsGood = TRUE ;
   switch ((char) curCh)
   { 
     case '+':  break ;
     case '0':
     case '1':
     case '2':
     case '3':
     case '4':
     case '5':
     case '6':
     case '7':
     case '8':
     case '9':  BackUpOneChar (cPtr, fPtr, "SynchOnFirstDig") ;  break ;
     default :  chIsGood = FALSE ;
    }
   return (chIsGood) ;
}
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */     
/*                       end file.c                          */
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */
