
/* BEGIN COPYRIGHT NOTICE

Copyright (c) 1992 by John Sarraille and Peter DiFalco
(john@ishi.csustan.edu)

Copyright (c) 1998 by John Sarraille
(john@ishi.csustan.edu)

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that copyright notice and this permission
notice appear in supporting documentation.

The algorithm used in this program was inspired by the paper
entitled "A Fast Algorithm To Determine Fractal Dimensions By
Box Counting", which was written by Liebovitch and Toth, and
which appeared in the journal "Physics Letters A", volume 141,
pp 386-390, (1989).

This program is not warranted: use at your own risk.

END COPYRIGHT NOTICE */

  /*

        HEADER FILE FOR FD3 SOFTWARE

        Filename:  qmacro.h

        THIS FILE CONTAINS MACRO VERSIONS OF FUNCTIONS IN THE
        FILE queue.c.  SEE THE DOCUMENTATION IN THAT FILE FOR
        INFORMATION ABOUT WHAT THESE MACROS DO.

  */


      /* THE MACROS IN THIS FILE: */

  /*

       EMPTYsTK
       PUSHsTK
       POPsTK
       EMPTYq
       ENQ
       DEQ
       TRANSFq

  */

/* ######################################################### */
/*                        EMPTYsTK                           */
/* ######################################################### */
         /*  MACRO VERSION OF EMPTY STACK TEST */

#define EMPTYsTK(stackPtr) (((stackPtr)->stkSize) == 0)

/* ######################################################### */
/*                         PUSHsTK                           */
/* ######################################################### */
         /*  MACRO VERSION OF STACK PUSH */

#define PUSHsTK(stackPtr, indexArray, eltIndex)            \
     ((indexArray)[(eltIndex)]) = ((stackPtr)->stkTop) ;   \
     ((stackPtr)->stkTop) = (eltIndex) ;                   \
     ((stackPtr)->stkSize)++ 

/* ######################################################### */
/*                          POPsTK                           */
/* ######################################################### */
         /*  MACRO VERSION OF STACK POP */

#define POPsTK(stackPtr, indexArray, eltIndexPtr)             \
   (*(eltIndexPtr)) = ((stackPtr)->stkTop) ;                  \
   ((stackPtr)->stkTop) = (indexArray)[(stackPtr)->stkTop] ;  \
   ((stackPtr)->stkSize)-- 

/* ######################################################### */
/*                         EMPTYq                            */
/* ######################################################### */
         /*  MACRO VERSION OF EMPTY QUEUE TEST */

#define EMPTYq(queuePtr) (((queuePtr)->Qsize) == 0)

/* ######################################################### */
/*                           ENQ                             */
/* ######################################################### */
         /*  MACRO VERSION OF ENQUEUE */

#define ENQ(queuePtr, indexArray, eltIndex)               \
  if  (((queuePtr)->Qsize) == 0)                          \
        ((queuePtr)->Qfront) = (eltIndex) ;               \
  else  (indexArray) [(queuePtr)->Qrear] = (eltIndex) ;   \
  ((queuePtr)->Qrear) = (eltIndex) ;                      \
  ((queuePtr)->Qsize)++ 

/* ######################################################### */
/*                         ENQfAST                           */
/* ######################################################### */
         /*  MACRO VERSION OF FAST ENQUEUE */

#define ENQfAST(queuePtr, indexArray, eltIndex)          \
  (indexArray) [(queuePtr)->Qrear] = (eltIndex) ;        \
  ((queuePtr)->Qrear) = (eltIndex) ;                     \
  ((queuePtr)->Qsize)++ 

/* ######################################################### */
/*                           DEQ                             */
/* ######################################################### */
         /*  MACRO VERSION OF DEQUEUE */

#define DEQ(queuePtr, indexArray, eltIndexPtr)              \
  *(eltIndexPtr) = ((queuePtr)->Qfront) ;                   \
  ((queuePtr)->Qfront) = (indexArray)[(queuePtr)->Qfront];  \
  ((queuePtr)->Qsize)-- 

/* ######################################################### */
/*                       DEQfAST                             */
/* ######################################################### */
         /*  MACRO VERSION OF FAST DEQUEUE */

#define DEQfAST(queuePtr, indexArray, eltIndexPtr)          \
  *(eltIndexPtr) = ((queuePtr)->Qfront) ;                   \
  ((queuePtr)->Qfront) = (indexArray)[(queuePtr)->Qfront];  \
  ((queuePtr)->Qsize)-- 

/* ######################################################### */
/*                         TRANSFq                           */
/* ######################################################### */
         /*  MACRO VERSION OF QUEUE TRANSFER */

#define TRANSFq(srcQueuePtr, targQueuePtr, indexArray)             \
  (indexArray) [(targQueuePtr)->Qrear] = ((srcQueuePtr)->Qfront) ; \
  ((targQueuePtr)->Qrear) = ((srcQueuePtr)->Qfront) ;              \
  ((srcQueuePtr)->Qfront) = (indexArray)[(srcQueuePtr)->Qfront];   \
  ((srcQueuePtr)->Qsize)-- ;                                       \
  ((targQueuePtr)->Qsize)++ 

/* ######################################################### */

/* ######################################################### */
/* ######################################################### */
/* ######################################################### */     
/*                       end qmacro.h                        */
/* ######################################################### */
/* ######################################################### */
/* ######################################################### */
